#include "server.h"
#include "types.h"
#include "view.h"
#include <stdlib.h>

static void xwayland_surface_map_notify(struct wl_listener* listener, void* data) {
    (void) data;

    /* Called when the surface is mapped, or ready to display on-screen. */
    flyaway_xwayland_view_t* xwayland_view = wl_container_of(listener, xwayland_view, map);
    xwayland_view->base->scene_tree = wlr_scene_subsurface_tree_create(&xwayland_view->base->server->scene->tree, xwayland_view->xwayland_surface->surface);
    
    flyaway_surface_t* surface = new_flyaway_surface_from_view(xwayland_view->base);
    xwayland_view->base->scene_tree->node.data = surface;
    xwayland_view->xwayland_surface->surface->data = surface;
    xwayland_view->base->surface = xwayland_view->xwayland_surface->surface;

    wlr_scene_node_set_position(&xwayland_view->base->scene_tree->node, xwayland_view->base->x, xwayland_view->base->y);
    
    wl_list_insert(&xwayland_view->base->server->views, &xwayland_view->base->link);
    focus_view(xwayland_view->base, xwayland_view->xwayland_surface->surface);
}

static void xwayland_surface_unmap_notify(struct wl_listener* listener, void* data) {
    (void) data;

    /* Called when the surface is unmapped, and should no longer be shown. */
    flyaway_xwayland_view_t* xwayland_view = wl_container_of(listener, xwayland_view, unmap);
    flyaway_server_t* server = xwayland_view->base->server;

    /* Reset the cursor mode if the grabbed view was unmapped. */
    if (xwayland_view->base == server->grabbed_view) {
        reset_cursor_mode(server);
    }

    if (server->seat->keyboard_state.focused_surface == xwayland_view->base->surface) {
        server->seat->keyboard_state.focused_surface = NULL;
    }

    wlr_scene_node_destroy(&xwayland_view->base->scene_tree->node);

    wl_list_remove(&xwayland_view->base->link);
}

static void xwayland_surface_destroy_notify(struct wl_listener* listener, void* data) {
    (void) data;

    /* Called when the surface is destroyed and should never be shown again. */
    flyaway_xwayland_view_t* xwayland_view = wl_container_of(listener, xwayland_view, destroy);

    wl_list_remove(&xwayland_view->map.link);
    wl_list_remove(&xwayland_view->unmap.link);
    wl_list_remove(&xwayland_view->destroy.link);

    free(xwayland_view);
}

static void xwayland_surface_request_configure_notify(struct wl_listener* listener, void* data) {
    flyaway_xwayland_view_t* xwayland_view = wl_container_of(listener, xwayland_view, request_configure);
    struct wlr_xwayland_surface* xwayland_surface = xwayland_view->xwayland_surface;
    struct wlr_xwayland_surface_configure_event* event = data;

    wlr_xwayland_surface_configure(xwayland_surface, event->x, event->y, event->width, event->height);
    xwayland_view->base->x = event->x;
    xwayland_view->base->y = event->y;

    if (xwayland_surface->mapped) {
        wlr_scene_node_set_position(&xwayland_view->base->scene_tree->node, event->x, event->y);
    }
}

static void xwayland_surface_set_geometry(struct wl_listener* listener, void* data) {
    (void) data;

    flyaway_xwayland_view_t* xwayland_view = wl_container_of(listener, xwayland_view, set_geometry);
    struct wlr_xwayland_surface* xwayland_surface = xwayland_view->xwayland_surface;
    xwayland_view->base->x = xwayland_surface->x;
    xwayland_view->base->y = xwayland_surface->y;
    if (xwayland_surface->mapped) {
        wlr_scene_node_set_position(&xwayland_view->base->scene_tree->node, xwayland_view->base->x, xwayland_view->base->y);
    }
}

flyaway_view_t* new_flyaway_xwayland_view(flyaway_server_t* server, struct wlr_xwayland_surface* xwayland_surface) {
    flyaway_view_t* view = calloc(1, sizeof(flyaway_xwayland_view_t));
    view->server = server;

    flyaway_xwayland_view_t* xwayland_view = calloc(1, sizeof(flyaway_xwayland_view_t));
    xwayland_view->base = view;
    xwayland_view->xwayland_surface = xwayland_surface;
    
    view->xwayland_view = xwayland_view;
    view->type = FLYAWAY_VIEW_TYPE_XWAYLAND;

    /* Listen to the various events it can emit */
    xwayland_view->map.notify = xwayland_surface_map_notify;
    wl_signal_add(&xwayland_surface->events.map, &xwayland_view->map);
    xwayland_view->unmap.notify = xwayland_surface_unmap_notify;
    wl_signal_add(&xwayland_surface->events.unmap, &xwayland_view->unmap);
    xwayland_view->destroy.notify = xwayland_surface_destroy_notify;
    wl_signal_add(&xwayland_surface->events.destroy, &xwayland_view->destroy);
    xwayland_view->request_configure.notify = xwayland_surface_request_configure_notify;
    wl_signal_add(&xwayland_surface->events.request_configure, &xwayland_view->request_configure);
    xwayland_view->set_geometry.notify = xwayland_surface_set_geometry;
    wl_signal_add(&xwayland_surface->events.set_geometry, &xwayland_view->set_geometry);

    return view;
}
